from __future__ import annotations

import argparse

from .count import generate_bootstrap_count_curves
from .count import generate_count_curve
from .froc import generate_bootstrap_froc_curves
from .froc import generate_froc_curve
from .utils import bounds
from .utils import test_point


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--bootstrap',
        type=int,
        default=1,
        help='Whether to do a single or bootstrap runs.',
    )
    parser.add_argument('--gt_ann', type=str, required=True)
    parser.add_argument('--pr_ann', type=str, required=True)
    parser.add_argument(
        '--use_iou',
        default=False,
        action='store_true',
        help='Use IoU score to decide based on `proximity`',
    )
    parser.add_argument(
        '--iou_thres',
        default=0.5,
        type=float,
        required=False,
        help='If IoU score is used the default threshold is set to .5',
    )
    parser.add_argument(
        '--n_sample_points',
        type=int,
        default=50,
        help='Number of points to evaluate the FROC curve at.',
    )
    parser.add_argument('--plot_title', type=str)
    parser.add_argument('--plot_output_path', type=str)

    parser.add_argument(
        '--test_ann',
        action='append',
        help='Extra ground-truth like annotations with annotator name/ID.',
        type=test_point,
        dest='test_ann',
        required=False,
    )

    parser.add_argument(
        '--counts',
        default=False,
        action='store_true',
    )

    parser.add_argument(
        '--weighted',
        default=False,
        action='store_true',
    )

    parser.add_argument(
        '--bounds',
        type=bounds,
        default=None,
        required=False,
    )

    parser.add_argument(
        '--csv',
        default=None,
        required=False,
        help='CSV file that contains alternative FROC data, for FROC curve generation ONLY',
    )

    args = parser.parse_args()

    if args.counts:
        if args.bootstrap > 1:
            generate_bootstrap_count_curves(
                gt_ann=args.gt_ann,
                pr_ann=args.pr_ann,
                n_bootstrap_samples=args.bootstrap,
                n_sample_points=args.n_sample_points,
                plot_title='Counts PR (bootstrap)' if args.plot_title is None else args.plot_title,
                plot_output_path='counts_bootstrap.png' if args.plot_output_path is None else args.plot_output_path,
                weighted=args.weighted,
                test_ann=args.test_ann,
                bounds=args.bounds,
            )
        else:
            generate_count_curve(
                gt_ann=args.gt_ann,
                pr_ann=args.pr_ann,
                weighted=args.weighted,
                plot_title='Counts PR' if args.plot_title is None else args.plot_title,
                plot_output_path='counts.png' if args.plot_output_path is None else args.plot_output_path,
                test_ann=args.test_ann,
                bounds=args.bounds,
            )

        exit(-1)

    if args.bootstrap > 1:
        print('Generating bootstrap curves... (this may take a while)')
        generate_bootstrap_froc_curves(
            gt_ann=args.gt_ann,
            pr_ann=args.pr_ann,
            n_bootstrap_samples=args.bootstrap,
            use_iou=args.use_iou,
            iou_thres=args.iou_thres,
            n_sample_points=args.n_sample_points,
            plot_title='FROC (bootstrap)' if args.plot_title is None else args.plot_title,
            plot_output_path='froc_bootstrap.png' if args.plot_output_path is None else args.plot_output_path,
            test_ann=args.test_ann,
            bounds=args.bounds,
            csv_path=args.csv,
        )
    else:
        print('Generating single FROC curve...')
        generate_froc_curve(
            gt_ann=args.gt_ann,
            pr_ann=args.pr_ann,
            use_iou=args.use_iou,
            iou_thres=args.iou_thres,
            n_sample_points=args.n_sample_points,
            plot_title='FROC' if args.plot_title is None else args.plot_title,
            plot_output_path='froc.png' if args.plot_output_path is None else args.plot_output_path,
            test_ann=args.test_ann,
            bounds=args.bounds,
            csv_path=args.csv,
        )



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import tqdm

from ..utils import build_gt_id2annotations
from ..utils import build_pr_id2annotations
from ..utils import COLORS
from ..utils import load_json_from_file
from ..utils import transform_gt_into_pr
from ..utils import update_scores
from .count_stats import init_stats
from .count_stats import update_stats


def count_point(gt_ann, pr_ann, score_thres, weighted):
    gt = load_json_from_file(gt_ann)
    pr = load_json_from_file(pr_ann)

    pr = update_scores(pr, score_thres)

    categories = gt['categories']

    stats = init_stats(gt, categories)

    gt_id_to_annotation = build_gt_id2annotations(gt)
    pr_id_to_annotation = build_pr_id2annotations(pr)

    stats = update_stats(
        stats, gt_id_to_annotation, pr_id_to_annotation,
        categories, weighted,
    )

    return stats


def calc_scores(stats, precision, recall):
    for category_id in stats:
        tp = stats[category_id]['TP']
        fp = stats[category_id]['FP']
        fn = stats[category_id]['FN']

        prec = tp / (tp + fp + 1e-7)
        rec = tp / (tp + fn + 1e-7)

        if precision.get(category_id, None) is None:
            precision[category_id] = []
        precision[category_id].append(prec)

        if recall.get(category_id, None) is None:
            recall[category_id] = []
        recall[category_id].append(rec)

    return precision, recall


def generate_count_curve(
    gt_ann,
    pr_ann,
    weighted=False,
    n_sample_points=50,
    plot_title='Count curve',
    plot_output_path='counts.png',
    test_ann=None,
    bounds=None,
):
    precision = {}
    recall = {}

    for score_thres in tqdm(
            np.linspace(0.0, 1.0, n_sample_points, endpoint=False),
    ):
        stats = count_point(gt_ann, pr_ann, score_thres, weighted)
        precision, recall = calc_scores(stats, precision, recall)

    if plot_title:
        fig, ax = plt.subplots(figsize=[27, 18])
        ins = ax.inset_axes([0.05, 0.05, 0.45, 0.4])
        ins.set_xticks(
            [.7, .75, .8, .85, .9, .95],
            [.7, .75, .8, .85, .9, .95], fontsize=30,
        )
        ins.yaxis.tick_right()
        ins.xaxis.tick_top()
        if bounds is not None:
            _, x_max, _, _ = bounds
            ins.set_xlim([.8, x_max])
        else:
            ins.set_xlim([.8, 1.0])

    for category_id in precision:
        prec = precision[category_id]
        rec = recall[category_id]
        if plot_title:
            ax.plot(
                rec,
                prec,
                'x--',
                label='AI ' + stats[category_id]['name'],
            )
            ins.plot(
                rec,
                prec,
                'x--',
                label='AI ' + stats[category_id]['name'],
            )

            if test_ann is not None:
                for t_ann, c in zip(test_ann, COLORS):
                    t_ann, label = t_ann
                    t_pr = transform_gt_into_pr(t_ann, gt_ann)
                    stats = count_point(gt_ann, t_pr, .5, weighted)
                    _precision, _recall = calc_scores(stats, {}, {})
                    if plot_title:
                        ax.plot(
                            _recall[category_id][0],
                            _precision[category_id][0],
                            'D',
                            markersize=15,
                            markeredgewidth=3,
                            label=label +
                            f' (R = {np.round(_recall[category_id][0], 3)})',
                            c=c,
                        )
                        ins.plot(
                            _recall[category_id][0],
                            _precision[category_id][0],
                            'D',
                            markersize=12,
                            markeredgewidth=2,
                            label=label +
                            f' (R = {np.round(_recall[category_id][0], 3)})',
                            c=c,
                        )
                        ax.hlines(
                            y=_precision[category_id][0],
                            xmin=np.min(rec),
                            xmax=np.max(rec),
                            linestyles='dashed',
                            colors=c,
                        )
                        ax.text(
                            x=_recall[category_id][0], y=_precision[category_id][0],
                            s=f' R = {np.round(_recall[category_id][0], 3)}',
                            fontdict={'fontsize': 20, 'fontweight': 'bold'},
                        )
                        ins.hlines(
                            y=_precision[category_id][0],
                            xmin=np.min(rec),
                            xmax=np.max(rec),
                            linestyles='dashed',
                            colors=c,
                        )

    if plot_title:
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        ax.legend(
            loc='lower left', bbox_to_anchor=(.65, .1),
            fancybox=True, shadow=True, ncol=1, fontsize=30,
        )

        ax.set_title(plot_title, fontdict={'fontsize': 35})
        ax.set_ylabel(
            'Precision', fontdict={
                'fontsize': 30,
            },
        )
        ax.set_xlabel('Recall', fontdict={'fontsize': 30})

        ax.tick_params(axis='both', which='major', labelsize=30)
        ins.tick_params(axis='both', which='major', labelsize=20)

        if bounds is not None:
            x_min, x_max, _, _ = bounds
            ax.set_xlim([x_min, x_max])
        else:
            ax.set_xlim([.7, 1.0])
            ax.set_ylim(bottom=0.05, top=1.02)
        fig.tight_layout()
        fig.savefig(plot_output_path, dpi=150)
    else:
        return precision, recall



from __future__ import annotations

import json
import os
import random
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import tqdm

from ..utils import COLORS
from ..utils import transform_gt_into_pr
from .count_curve import calc_scores
from .count_curve import count_point
from .count_curve import generate_count_curve


def generate_bootstrap_count_curves(
    gt_ann,
    pr_ann,
    weighted=False,
    n_sample_points=50,
    n_bootstrap_samples=5,
    plot_title='Count curve',
    plot_output_path='counts.png',
    test_ann=None,
    bounds=None,
):
    with open(gt_ann) as fp:
        GT_ANN = json.load(fp)

    with open(pr_ann) as fp:
        PRED_ANN = json.load(fp)

    n_images = len(GT_ANN['images'])

    fig, ax = plt.subplots(figsize=[27, 18])
    ins = ax.inset_axes([0.05, 0.05, 0.45, 0.4])
    ins.set_xticks([.85, .9, .95], [.85, .9, .95], fontsize=40)
    ins.yaxis.tick_right()
    ins.xaxis.tick_top()

    if bounds is not None:
        _, x_max, _, _ = bounds
        ins.set_xlim([.8, x_max])
    else:
        ins.set_xlim([.8, 1.0])

    collected_rocs = {'precision': {}, 'recall': {}}

    _, non_bootstrap_rec = generate_count_curve(
        gt_ann,
        pr_ann,
        weighted=weighted,
        n_sample_points=n_sample_points,
        plot_title=None,
        plot_output_path=None,
    )

    for _ in tqdm(range(n_bootstrap_samples)):
        selected_images = random.choices(GT_ANN['images'], k=n_images)
        bootstrap_gt = deepcopy(GT_ANN)

        del bootstrap_gt['images']

        bootstrap_gt['images'] = selected_images

        gt_annotations = bootstrap_gt['annotations']

        del bootstrap_gt['annotations']

        bootstrap_gt['annotations'] = []
        for _gt_ann_ in gt_annotations:
            img_id = _gt_ann_['image_id']
            for ind, selected_image in enumerate(selected_images):
                if selected_image['id'] == img_id:
                    new_gt_ann = deepcopy(_gt_ann_)
                    new_gt_ann['image_id'] = ind
                    bootstrap_gt['annotations'].append(new_gt_ann)

        predictions = []

        for ind, img in enumerate(selected_images):
            for pr in PRED_ANN:
                if pr['image_id'] == img['id']:
                    new_pr = deepcopy(pr)
                    new_pr['image_id'] = ind
                    predictions.append(new_pr)

        re_indexed_images = []
        for ind in range(len(selected_images)):
            image = deepcopy(selected_images[ind])
            image['id'] = ind
            re_indexed_images.append(image)

        bootstrap_gt['images'] = re_indexed_images

        with open('/tmp/tmp_bootstrap_gt.json', 'w') as fp:
            json.dump(bootstrap_gt, fp)

        with open('/tmp/tmp_bootstrap_pred.json', 'w') as fp:
            json.dump(predictions, fp)

        tmp_gt_ann = '/tmp/tmp_bootstrap_gt.json'
        tmp_pred_ann = '/tmp/tmp_bootstrap_pred.json'

        precision, recall = generate_count_curve(
            tmp_gt_ann,
            tmp_pred_ann,
            weighted=weighted,
            n_sample_points=n_sample_points,
            plot_title=None,
            plot_output_path=None,
        )

        for cat_id in precision:
            if collected_rocs['precision'].get(cat_id, None) is None:
                collected_rocs['precision'] = {cat_id: []}
            if collected_rocs['recall'].get(cat_id, None) is None:
                collected_rocs['recall'] = {cat_id: []}

        for cat_id in precision:
            collected_rocs['precision'][cat_id].append(precision[cat_id])
            collected_rocs['recall'][cat_id].append(recall[cat_id])

    interpolated_rocs = {}
    max_roc_prec = {}
    min_roc_prec = {}

    for cat_id in collected_rocs['precision']:
        all_prec = np.array(collected_rocs['precision'][cat_id]).reshape(
            n_bootstrap_samples, n_sample_points,
        )
        all_rec = np.array(collected_rocs['recall'][cat_id]).reshape(
            n_bootstrap_samples, n_sample_points,
        )

        min_rec, max_rec = np.min(non_bootstrap_rec[cat_id]), np.max(
            non_bootstrap_rec[cat_id],
        )

        if bounds is not None:
            min_rec, max_rec = bounds[0], bounds[1]

        x_range = np.linspace(min_rec, max_rec, n_sample_points, endpoint=True)

        rocs = []

        for prec, rec in zip(all_prec, all_rec):
            interpolated_prec = np.interp(x_range, rec[::-1], prec[::-1])
            rocs.append(interpolated_prec)

        interpolated_rocs[cat_id] = {
            'rec': x_range,
            'prec': np.array(rocs).reshape(
                n_bootstrap_samples,
                n_sample_points,
            ),
        }

        max_roc_prec[cat_id] = np.max(
            np.array(rocs).reshape(n_bootstrap_samples, n_sample_points),
            axis=0,
        )

        min_roc_prec[cat_id] = np.min(
            np.array(rocs).reshape(n_bootstrap_samples, n_sample_points),
            axis=0,
        )

    mean_roc_curve = {}

    for cat_id in interpolated_rocs:
        mean_roc_curve[cat_id] = np.stack(
            (
                interpolated_rocs[cat_id]['rec'],
                np.mean(interpolated_rocs[cat_id]['prec'], axis=0),
            ),
            axis=-1,
        )

        ax.plot(
            mean_roc_curve[cat_id][:, 0],
            mean_roc_curve[cat_id][:, 1],
            'b-',
            label='mean',
        )

        ins.plot(
            mean_roc_curve[cat_id][:, 0],
            mean_roc_curve[cat_id][:, 1],
            'b-',
            label='mean',
        )

        ax.fill_between(
            interpolated_rocs[cat_id]['rec'],
            min_roc_prec[cat_id],
            max_roc_prec[cat_id],
            alpha=.2,
        )

        ins.fill_between(
            interpolated_rocs[cat_id]['rec'],
            min_roc_prec[cat_id],
            max_roc_prec[cat_id],
            alpha=.2,
        )

        if test_ann is not None:
            for t_ann, c in zip(test_ann, COLORS):
                t_ann, label = t_ann
                t_pr = transform_gt_into_pr(t_ann, gt_ann)
                stats = count_point(gt_ann, t_pr, .5, weighted)
                _prec_accuracy, _rec_per_image = calc_scores(stats, {}, {})
                ax.plot(
                    _rec_per_image[cat_id][0],
                    _prec_accuracy[cat_id][0],
                    'D',
                    markersize=15,
                    markeredgewidth=3,
                    label=label +
                    f' (R = {np.round(_rec_per_image[cat_id][0], 3)})',
                    c=c,
                )
                ins.plot(
                    _rec_per_image[cat_id][0],
                    _prec_accuracy[cat_id][0],
                    'D',
                    markersize=15,
                    markeredgewidth=3,
                    label=label +
                    f' (R = {np.round(_rec_per_image[cat_id][0], 3)})',
                    c=c,
                )

                if bounds is not None:
                    min_rec, max_rec = bounds[0], bounds[1]
                    if _rec_per_image[cat_id][0] < min_rec:
                        continue
                else:
                    ax.text(
                        x=_rec_per_image[cat_id][0], y=_prec_accuracy[cat_id][0],
                        s=f' R = {np.round(_rec_per_image[cat_id][0], 3)}',
                        fontdict={'fontsize': 35, 'fontweight': 'bold'},
                    )

                ax.hlines(
                    y=_prec_accuracy[cat_id][0],
                    xmin=min_rec,
                    xmax=max_rec,
                    linestyles='dashed',
                    colors=c,
                )

                ins.hlines(
                    y=_prec_accuracy[cat_id][0],
                    xmin=min_rec,
                    xmax=max_rec,
                    linestyles='dashed',
                    colors=c,
                )

    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    ax.legend(
        loc='lower left', bbox_to_anchor=(.65, .1),
        fancybox=True, shadow=True, ncol=1, fontsize=40,
    )

    ax.set_title(plot_title, fontdict={'fontsize': 45})
    ax.set_ylabel(
        'Precision', fontdict={
            'fontsize': 45,
        },
    )
    ax.set_xlabel('Recall', fontdict={'fontsize': 45})

    ax.tick_params(axis='both', which='major', labelsize=40)
    ins.tick_params(axis='both', which='major', labelsize=40)

    if bounds is not None:
        x_min, x_max, _, _ = bounds
        ax.set_xlim([x_min, x_max])
    else:
        ax.set_xlim([.7, 1.0])
        ax.set_ylim(bottom=0.05, top=1.02)

    ax.grid(True, which='both', axis='both', alpha=.5, linestyle='--')
    fig.tight_layout()
    fig.savefig(plot_output_path, dpi=150)

    os.remove('/tmp/tmp_bootstrap_gt.json')
    os.remove('/tmp/tmp_bootstrap_pred.json')



from __future__ import annotations

import numpy as np
from scipy.optimize import linear_sum_assignment

from ..utils import get_overlap


def init_stats(gt: dict, categories: dict) -> dict:
    stats = {
        cat['id']: {
            'name': cat['name'],
            'P': 0,
            'TP': 0,
            'FP': 0,
            'FN': 0,
        }
        for cat in categories
    }

    return stats


def update_stats(
    stats: dict, gt_id_to_annotation: dict,
    pr_id_to_annotation: dict, categories: dict, weighted: bool,
):
    for image_id in gt_id_to_annotation:
        cat2anns: dict[int, dict[str, list]] = {}
        for cat in categories:
            cat2anns[cat['id']] = {'gt': [], 'pr': []}

        for gt_ann in gt_id_to_annotation[image_id]:
            cat2anns[gt_ann['category_id']]['gt'].append(gt_ann)
        for pred_ann in pr_id_to_annotation.get(image_id, []):
            cat2anns[pred_ann['category_id']]['pr'].append(pred_ann)

        for cat in categories:
            gt_anns = cat2anns[cat['id']]['gt']
            pr_anns = cat2anns[cat['id']]['pr']

            n_gt = len(gt_anns)
            n_pr = len(pr_anns)

            if n_gt == 0:
                if n_pr == 0:
                    continue
                else:
                    stats[cat['id']]['FP'] += n_pr
            else:
                cost_matrix = np.ones((n_gt, n_pr)) * np.finfo(np.float64).max

                for gt_ind, gt_ann in enumerate(gt_anns):
                    for pr_ind, pr_ann in enumerate(pr_anns):
                        if weighted:
                            overlap = get_overlap(
                                gt_ann['bbox'],
                                pr_ann['bbox'],
                            )
                            if overlap > 0.:
                                weight = 1. / (
                                    overlap +
                                    np.random.uniform(0, 1) / 1e-6
                                )
                                cost_matrix[gt_ind, pr_ind] = weight
                        else:
                            gt_x, gt_y, gt_w, gt_h = gt_ann['bbox']

                            pr_x, pr_y, pr_w, pr_h = pr_ann['bbox']
                            pr_bbox_center = pr_x + pr_w / 2, pr_y + pr_h / 2

                            if (
                                pr_bbox_center[0] >= gt_x
                                and pr_bbox_center[0] <= gt_x + gt_w
                                and pr_bbox_center[1] >= gt_y
                                and pr_bbox_center[1] <= gt_y + gt_h
                            ):
                                cost_matrix[
                                    gt_ind,
                                    pr_ind,
                                ] = 1.0  # connected, not weighted

                row_ind, col_ind = linear_sum_assignment(
                    cost_matrix,
                )  # Hungarian-matching

                n_true_positives = len(row_ind)
                n_false_positives = max(n_pr - len(col_ind), 0)
                n_false_negatives = max(n_gt - len(row_ind), 0)

                stats[cat['id']]['P'] += n_gt
                stats[cat['id']]['TP'] += n_true_positives
                stats[cat['id']]['FP'] += n_false_positives
                stats[cat['id']]['FN'] += n_false_negatives

    return stats



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import tqdm

from ..utils import build_gt_id2annotations
from ..utils import build_pr_id2annotations
from ..utils import COLORS
from ..utils import load_json_from_file
from ..utils import transform_gt_into_pr
from ..utils import update_scores
from .froc_stats import init_stats
from .froc_stats import update_stats

import pandas as pd


def froc_point(gt_ann, pr_ann, score_thres, use_iou, iou_thres):
    gt = load_json_from_file(gt_ann)
    pr = load_json_from_file(pr_ann)

    pr = update_scores(pr, score_thres)

    categories = gt['categories']

    stats = init_stats(gt, categories)

    gt_id_to_annotation = build_gt_id2annotations(gt)
    pr_id_to_annotation = build_pr_id2annotations(pr)

    stats = update_stats(
        stats, gt_id_to_annotation, pr_id_to_annotation,
        categories, use_iou, iou_thres,
    )

    return stats


def calc_scores(stats, lls_accuracy, nlls_per_image):
    for category_id in stats:
        if lls_accuracy.get(category_id, None):
            lls_accuracy[category_id].append(
                stats[category_id]['LL'] /
                stats[category_id]['n_lesions'],
            )
        else:
            lls_accuracy[category_id] = []
            lls_accuracy[category_id].append(
                stats[category_id]['LL'] /
                stats[category_id]['n_lesions'],
            )

        if nlls_per_image.get(category_id, None):
            nlls_per_image[category_id].append(
                stats[category_id]['NL'] /
                stats[category_id]['n_images'],
            )
        else:
            nlls_per_image[category_id] = []
            nlls_per_image[category_id].append(
                stats[category_id]['NL'] /
                stats[category_id]['n_images'],
            )

    return lls_accuracy, nlls_per_image


def generate_froc_curve(
    gt_ann,
    pr_ann,
    use_iou=False,
    iou_thres=0.5,
    n_sample_points=50,
    plot_title='FROC curve',
    plot_output_path='froc.png',
    test_ann=None,
    bounds=None,
    csv_path=None,
):

    lls_accuracy = {}
    nlls_per_image = {}

    for score_thres in tqdm(
            np.linspace(0.0, 1.0, n_sample_points, endpoint=False),
    ):
        stats = froc_point(gt_ann, pr_ann, score_thres, use_iou, iou_thres)
        lls_accuracy, nlls_per_image = calc_scores(
            stats, lls_accuracy,
            nlls_per_image,
        )

    if plot_title:
        fig, ax = plt.subplots(figsize=[27, 18])
        ins = ax.inset_axes([0.55, 0.05, 0.45, 0.4])
        ins.set_xticks(
            [0.1, 1.0, 2.0, 3.0, 4.0], [
                0.1, 1.0, 2.0, 3.0, 4.0,
            ], fontsize=30,
        )

        if bounds is not None:
            _, x_max, _, y_max = bounds
            ins.set_xlim([.1, x_max])
        else:
            ins.set_xlim([0.1, 4.5])

    for category_id in lls_accuracy:
        lls = lls_accuracy[category_id]
        nlls = nlls_per_image[category_id]
        if plot_title:
            ax.semilogx(
                nlls,
                lls,
                'x--',
                label='AI ' + stats[category_id]['name'],
            )
            ins.plot(
                nlls,
                lls,
                'x--',
                label='AI ' + stats[category_id]['name'],
            )

            if test_ann is not None:
                for t_ann, c in zip(test_ann, COLORS):
                    t_ann, label = t_ann
                    t_pr = transform_gt_into_pr(t_ann, gt_ann)
                    stats = froc_point(gt_ann, t_pr, .5, use_iou, iou_thres)
                    _lls_accuracy, _nlls_per_image = calc_scores(stats, {}, {})
                    if plot_title:
                        ax.plot(
                            _nlls_per_image[category_id][0],
                            _lls_accuracy[category_id][0],
                            'D',
                            markersize=15,
                            markeredgewidth=3,
                            label=label +
                            f' (FP/image = {np.round(_nlls_per_image[category_id][0], 2)})',
                            c=c,
                        )
                        ins.plot(
                            _nlls_per_image[category_id][0],
                            _lls_accuracy[category_id][0],
                            'D',
                            markersize=12,
                            markeredgewidth=2,
                            label=label +
                            f' (FP/image = {np.round(_nlls_per_image[category_id][0], 2)})',
                            c=c,
                        )
                        ax.hlines(
                            y=_lls_accuracy[category_id][0],
                            xmin=np.min(nlls),
                            xmax=np.max(nlls),
                            linestyles='dashed',
                            colors=c,
                        )
                        ins.hlines(
                            y=_lls_accuracy[category_id][0],
                            xmin=np.min(nlls),
                            xmax=np.max(nlls),
                            linestyles='dashed',
                            colors=c,
                        )
                        ax.text(
                            x=_nlls_per_image[category_id][0], y=_lls_accuracy[category_id][0],
                            s=f' FP/image = {np.round(_nlls_per_image[category_id][0], 2)}',
                            fontdict={'fontsize': 20, 'fontweight': 'bold'},
                        )
                        
    if csv_path is not None and plot_title is not None:
        alternative_froc = pd.read_csv(csv_path)
        alternative_ll, alternative_nll = alternative_froc['lls'], alternative_froc['nlls']
        alternative_ll_high, alternative_nll_high = alternative_froc['lls_high'], alternative_froc['nlls_high']
        
        alternative_ll = np.interp(alternative_nll_high, alternative_nll, alternative_ll)
        alternative_ll_low = alternative_ll - (alternative_ll_high - alternative_ll)
        
        ax.semilogx(alternative_nll_high, alternative_ll, 'gx--', label='Alternative FROC')
        ins.semilogx(alternative_nll_high, alternative_ll, 'gx--', label='Alternative FROC')
        
        ax.fill_between(alternative_nll_high, alternative_ll_low, alternative_ll_high, alpha=.2)
        ins.fill_between(alternative_nll_high, alternative_ll_low, alternative_ll_high, alpha=.2)

    if plot_title:
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        ax.legend(
            loc='lower left', bbox_to_anchor=(.1, .1),
            fancybox=True, shadow=True, ncol=1, fontsize=30,
        )

        ax.set_title(plot_title, fontdict={'fontsize': 35})
        ax.set_ylabel('Sensitivity', fontdict={'fontsize': 30})
        ax.set_xlabel('FP / image', fontdict={'fontsize': 30})

        ax.tick_params(axis='both', which='major', labelsize=30)
        ins.tick_params(axis='both', which='major', labelsize=20)

        if bounds is not None:
            x_min, x_max, y_min, y_max = bounds
            ax.set_ylim([y_min, y_max])
            ax.set_xlim([x_min, x_max])
        else:
            ax.set_ylim(bottom=0.05, top=1.02)
        
        if csv_path is not None:
            min_x, max_x = min(alternative_nll_high), max(alternative_nll_high)
            ax.set_xlim([min_x, max_x])
            
        fig.tight_layout()
        fig.savefig(fname=plot_output_path, dpi=150)
    else:
        return lls_accuracy, nlls_per_image



from __future__ import annotations

import numpy as np
from scipy.optimize import linear_sum_assignment

from ..utils import get_iou_score


def init_stats(gt: dict, categories: dict) -> dict:
    """Initializing the statistics before counting leasion
       and non-leasion localiazations.

    Arguments:
        gt {dict} -- Ground truth COCO dataset
        categories {dict} -- Dictionary of categories in the COCO dataset

    Returns:
        stats {dict} -- Statistics to be updated, containing every information
                        necessary to evaluate a single FROC point
    """
    stats = {
        cat['id']: {
            'name': cat['name'],
            'LL': 0,
            'NL': 0,
            'n_images': [],
            'n_lesions': 0,
        }
        for cat in categories
    }

    for annotation in gt['annotations']:
        category_id = annotation['category_id']
        stats[category_id]['n_lesions'] += 1

    for image in gt['images']:
        image_id = image['id']
        for cat_id in stats:
            stats[cat_id]['n_images'].append(image_id)

    for cat_id in stats:
        stats[cat_id]['n_images'] = len(stats[cat_id]['n_images'])

    return stats


def update_stats(
    stats: dict,
    gt_id_to_annotation: dict,
    pr_id_to_annotation: dict,
    categories: dict,
    use_iou: bool,
    iou_thres: float,
):
    """Updating statistics as going through images of the dataset.

    Arguments:
        stats {dict} -- FROC statistics
        gt_id_to_annotation {dict} -- Ground-truth image IDs to annotations.
        pr_id_to_annotation {dict} -- Prediction image IDs to annotations.
        categories {dict} -- COCO categories dictionary.
        use_iou {bool} -- Whether or not to use iou thresholding.
        iou_thres {float} -- IoU threshold when using IoU thresholding.

    Returns:
        stats {dict} -- Updated FROC statistics
    """
    for image_id in gt_id_to_annotation:
        cat2anns: dict[int, dict[str, list]] = {}
        for cat in categories:
            cat2anns[cat['id']] = {'gt': [], 'pr': []}

        for gt_ann in gt_id_to_annotation[image_id]:
            cat2anns[gt_ann['category_id']]['gt'].append(gt_ann)
        for pred_ann in pr_id_to_annotation.get(image_id, []):
            cat2anns[pred_ann['category_id']]['pr'].append(pred_ann)

        for cat in categories:
            gt_anns = cat2anns[cat['id']]['gt']
            pr_anns = cat2anns[cat['id']]['pr']

            n_gt = len(gt_anns)
            n_pr = len(pr_anns)

            if n_gt == 0:
                if n_pr == 0:
                    continue
                stats[cat['id']]['NL'] += n_pr
            else:
                cost_matrix = np.ones((n_gt, n_pr)) * 1e6

                for gt_ind, gt_ann in enumerate(gt_anns):
                    for pr_ind, pr_ann in enumerate(pr_anns):
                        if use_iou:
                            iou_score = get_iou_score(
                                gt_ann['bbox'],
                                pr_ann['bbox'],
                            )
                            if iou_score > iou_thres:
                                cost_matrix[gt_ind, pr_ind] = iou_score / (
                                    np.random.uniform(0, 1) / 1e6
                                )
                        else:
                            gt_x, gt_y, gt_w, gt_h = gt_ann['bbox']

                            pr_x, pr_y, pr_w, pr_h = pr_ann['bbox']
                            pr_bbox_center = pr_x + pr_w / 2, pr_y + pr_h / 2

                            if (
                                pr_bbox_center[0] >= gt_x
                                and pr_bbox_center[0] <= gt_x + gt_w
                                and pr_bbox_center[1] >= gt_y
                                and pr_bbox_center[1] <= gt_y + gt_h
                            ):
                                cost_matrix[gt_ind, pr_ind] = 1.0

                row_ind, col_ind = linear_sum_assignment(
                    cost_matrix,
                )  # Hungarian-matching

                n_true_positives = len(row_ind)
                n_false_positives = max(n_pr - len(col_ind), 0)

                stats[cat['id']]['LL'] += n_true_positives
                stats[cat['id']]['NL'] += n_false_positives

    return stats



from __future__ import annotations

from .bootstrap_count_curve import generate_bootstrap_count_curves  # noqa: F401, E501
from .count_curve import count_point  # noqa: F401
from .count_curve import generate_count_curve  # noqa: F401



from __future__ import annotations

from .bootstrap_froc_curve import generate_bootstrap_froc_curves  # noqa: F401, E501
from .froc_curve import froc_point  # noqa: F401
from .froc_curve import generate_froc_curve  # noqa: F401



from __future__ import annotations

import argparse
import json
import random
import string
from pathlib import Path

COLORS = [
    'r', 'g', 'm', 'y', 'teal', 'magenta',
    'royalblue', 'lime', 'forestgreen',
]


def test_point(val):
    try:
        json_file, annotator = val.split(',')
        return json_file, annotator
    except ValueError:
        print('assigning random annotator ID')
        letters = string.ascii_lowercase
        return val, 'test_' + ''.join(random.choice(letters) for _ in range(3))


def bounds(val):
    try:
        x_min, x_max, y_min, y_max = val.split(',')
        return float(x_min), float(x_max), float(y_min), float(y_max)
    except ValueError:
        raise argparse.ArgumentTypeError(
            'axis bounds must be four comma separated floats, (x_min, x_max, y_min, y_max)',
        )


def transform_gt_into_pr(
    gt: dict[str, list],
    true_gt: dict[str, list],
) -> str:
    gt = load_json_from_file(gt)
    true_gt = load_json_from_file(true_gt)
    id2trueId = {}
    for true_gt_img in true_gt['images']:
        for gt_img in gt['images']:
            true_gt_img_name = true_gt_img['file_name']
            gt_img_name = gt_img['file_name']
            if true_gt_img_name == gt_img_name:
                id2trueId[gt_img['id']] = true_gt_img['id']
    annotations = gt['annotations']
    scored_annotations = []
    for anno in annotations:
        if id2trueId.get(anno['image_id']) is not None:
            anno['image_id'] = id2trueId[anno['image_id']]
            anno['score'] = 1.0
            anno['category_id'] = 0
            scored_annotations.append(anno)
    with open('/tmp/pr2gt.json', 'w') as fp:
        json.dump(scored_annotations, fp)
    return '/tmp/pr2gt.json'


def load_json_from_file(file_path):
    with Path(file_path).open() as fp:
        data = json.load(fp)
    return data


def update_scores(json_data: list, score_thres: float) -> list:
    preds = []
    for _, pred in enumerate(json_data):
        if pred['score'] > score_thres:
            preds.append(pred)
    return preds


def get_overlap(gt_box: list, pr_box: list) -> float:
    """Intersection score between GT and prediction boxes.

    Arguments:
        gt_box {list} -- [x, y, w, h] of ground-truth lesion
        pr_box {list} -- [x, y, w, h] of prediction bounding box

    Returns:
        intersection {float}
    """
    gt_x, gt_y, gt_w, gt_h = gt_box
    pr_x, pr_y, pr_w, pr_h = pr_box

    xA = max(gt_x, pr_x)
    xB = min(gt_x + gt_w, pr_x + pr_w)
    yA = max(gt_y, pr_y)
    yB = min(gt_y + gt_h, pr_y + pr_h)

    return float(max((xB - xA), 0) * max((yB - yA), 0))


def get_iou_score(gt_box: list, pr_box: list) -> float:
    """IoU score between GT and prediction boxes.

    Arguments:
        gt_box {list} -- [x, y, w, h] of ground-truth lesion
        pr_box {list} -- [x, y, w, h] of prediction bounding box

    Returns:
        score {float} -- intersection over union score of the two boxes
    """
    *_, gt_w, gt_h = gt_box
    *_, pr_w, pr_h = pr_box

    intersection = get_overlap(gt_box, pr_box)

    gt_area = gt_w * gt_h
    pr_area = pr_w * pr_h

    return intersection / (gt_area + pr_area - intersection)


def build_pr_id2annotations(pr: list) -> dict:
    """Build image to annotation dictionary based on list of predictions.

    Arguments:
        pr {list} -- Predictions from Detectron2 (coco-instance-results.json)

    Returns:
        id_to_annotation {dict} -- Image IDs to predicted annotations.
    """
    id_to_annotation: dict[int, list[str]] = dict()
    for annotation in pr:
        id_to_annotation[annotation['image_id']] = []
    for annotation in pr:
        # DO not add the exact same annotation twice in the bootstrap case
        for ann in id_to_annotation[annotation['image_id']]:
            if ann == annotation:
                continue
        id_to_annotation[annotation['image_id']].append(annotation)
    return id_to_annotation


def build_gt_id2annotations(gt: dict) -> dict:
    """Build image to annotation dictionary based on the ground truth dataset.

    Arguments:
        gt {dict} -- COCO dataset of ground truth annotations.

    Returns:
        id_to_annotation {dict} -- Image IDs to ground-truth annotations.
    """
    id_to_annotation: dict[int, list[str]] = dict()
    for image in gt['images']:
        id_to_annotation[image['id']] = []
    for annotation in gt['annotations']:
        # DO not add the exact same annotation twice in the bootstrap case
        for ann in id_to_annotation[annotation['image_id']]:
            if ann == annotation:
                continue
        id_to_annotation[annotation['image_id']].append(annotation)
    return id_to_annotation



from __future__ import annotations

import json
import os
import random
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import tqdm

from ..utils import transform_gt_into_pr
from .froc_curve import calc_scores
from .froc_curve import COLORS
from .froc_curve import froc_point
from .froc_curve import generate_froc_curve

import pandas as pd


def generate_bootstrap_froc_curves(
    gt_ann,
    pr_ann,
    n_bootstrap_samples=5,
    use_iou=False,
    iou_thres=0.5,
    n_sample_points=50,
    plot_title='Bootstrap FROC',
    plot_output_path='froc_bootstrapped.png',
    test_ann=None,
    bounds=None,
    csv_path=None,
):
    with open(gt_ann) as fp:
        GT_ANN = json.load(fp)

    with open(pr_ann) as fp:
        PRED_ANN = json.load(fp)

    n_images = len(GT_ANN['images'])

    fig, ax = plt.subplots(figsize=[27, 18])
    ins = ax.inset_axes([0.55, 0.05, 0.45, 0.4])
    ins.set_xticks(
        [0.1, 1.0, 2.0, 3.0, 4.0], [
            0.1, 1.0, 2.0, 3.0, 4.0,
        ], fontsize=45,
    )

    ins.set_xlim([.1, 4.5])

    collected_frocs = {'lls': {}, 'nlls': {}}

    non_bootstrap_lls, non_bootstrap_nlls = generate_froc_curve(
        gt_ann,
        pr_ann,
        use_iou,
        iou_thres,
        n_sample_points,
        plot_title=None,
        plot_output_path=None,
    )

    for _ in tqdm(range(n_bootstrap_samples)):
        selected_images = random.choices(
            GT_ANN['images'], k=n_images,
        )  # sample with replacement
        bootstrap_gt = deepcopy(GT_ANN)

        del bootstrap_gt['images']

        bootstrap_gt['images'] = selected_images

        gt_annotations = bootstrap_gt['annotations']

        del bootstrap_gt['annotations']

        bootstrap_gt['annotations'] = []
        for _gt_ann_ in gt_annotations:
            img_id = _gt_ann_['image_id']
            for ind, selected_image in enumerate(selected_images):
                if selected_image['id'] == img_id:
                    new_gt_ann = deepcopy(_gt_ann_)
                    new_gt_ann['image_id'] = ind
                    bootstrap_gt['annotations'].append(new_gt_ann)

        predictions = []

        for ind, img in enumerate(selected_images):
            for pr in PRED_ANN:
                if pr['image_id'] == img['id']:
                    new_pr = deepcopy(pr)
                    new_pr['image_id'] = ind
                    predictions.append(new_pr)

        re_indexed_images = []
        for ind in range(len(selected_images)):
            image = deepcopy(selected_images[ind])
            image['id'] = ind
            re_indexed_images.append(image)

        bootstrap_gt['images'] = re_indexed_images

        with open('/tmp/tmp_bootstrap_gt.json', 'w') as fp:
            json.dump(bootstrap_gt, fp)

        with open('/tmp/tmp_bootstrap_pred.json', 'w') as fp:
            json.dump(predictions, fp)

        tmp_gt_ann = '/tmp/tmp_bootstrap_gt.json'
        tmp_pred_ann = '/tmp/tmp_bootstrap_pred.json'

        lls, nlls = generate_froc_curve(
            tmp_gt_ann,
            tmp_pred_ann,
            use_iou,
            iou_thres,
            n_sample_points,
            plot_title=None,
            plot_output_path=None,
        )

        for cat_id in lls:
            if collected_frocs['lls'].get(cat_id, None) is None:
                collected_frocs['lls'] = {cat_id: []}
            if collected_frocs['nlls'].get(cat_id, None) is None:
                collected_frocs['nlls'] = {cat_id: []}

        for cat_id in lls:
            collected_frocs['lls'][cat_id].append(lls[cat_id])
            collected_frocs['nlls'][cat_id].append(nlls[cat_id])

    interpolated_frocs = {}
    max_froc_lls = {}
    min_froc_lls = {}

    min_nlls, max_nlls = np.min(non_bootstrap_nlls[cat_id]), np.max(
        non_bootstrap_nlls[cat_id],
    )

    if bounds is not None:
        min_nlls, max_nlls = bounds[0], bounds[1]

    x_range = np.logspace(
        np.log10(min_nlls + 1e-8), np.log10(max_nlls),
        n_sample_points, endpoint=True,
    )
    
    if csv_path is not None:
        alternative_froc = pd.read_csv(csv_path)
        alternative_ll, alternative_nll = alternative_froc['lls'], alternative_froc['nlls']
        alternative_ll_high, alternative_nll_high = alternative_froc['lls_high'], alternative_froc['nlls_high']
        
        alternative_ll = np.interp(alternative_nll_high, alternative_nll, alternative_ll)
        alternative_ll_low = alternative_ll - (alternative_ll_high - alternative_ll)
        
        ax.semilogx(alternative_nll_high, alternative_ll, 'gx--', label='Alternative FROC')
        ins.semilogx(alternative_nll_high, alternative_ll, 'gx--', label='Alternative FROC')
        
        ax.fill_between(alternative_nll_high, alternative_ll_low, alternative_ll_high, alpha=.2)
        ins.fill_between(alternative_nll_high, alternative_ll_low, alternative_ll_high, alpha=.2)

    for cat_id in collected_frocs['lls']:
        all_lls = np.array(collected_frocs['lls'][cat_id]).reshape(
            n_bootstrap_samples, n_sample_points,
        )
        all_nlls = np.array(collected_frocs['nlls'][cat_id]).reshape(
            n_bootstrap_samples, n_sample_points,
        )

        frocs = []

        for lls, nlls in zip(all_lls, all_nlls):
            interpolated_lls = np.interp(x_range, nlls[::-1], lls[::-1])
            frocs.append(interpolated_lls)

        interpolated_frocs[cat_id] = {
            'nlls': x_range,
            'lls': np.array(frocs).reshape(
                n_bootstrap_samples,
                n_sample_points,
            ),
        }

        max_froc_lls[cat_id] = np.max(
            np.array(frocs).reshape(n_bootstrap_samples, n_sample_points),
            axis=0,
        )

        min_froc_lls[cat_id] = np.min(
            np.array(frocs).reshape(n_bootstrap_samples, n_sample_points),
            axis=0,
        )

    for cat_id in interpolated_frocs:
        ax.semilogx(
            interpolated_frocs[cat_id]['nlls'],
            np.mean(interpolated_frocs[cat_id]['lls'], axis=0),
            'b-',
            label='mean',
        )

        ins.plot(
            interpolated_frocs[cat_id]['nlls'],
            np.mean(interpolated_frocs[cat_id]['lls'], axis=0),
            'b-',
            label='mean',
        )

        ax.fill_between(
            interpolated_frocs[cat_id]['nlls'],
            min_froc_lls[cat_id],
            max_froc_lls[cat_id],
            alpha=.2,
        )

        ins.fill_between(
            interpolated_frocs[cat_id]['nlls'],
            min_froc_lls[cat_id],
            max_froc_lls[cat_id],
            alpha=.2,
        )

        if test_ann is not None:
            for t_ann, c in zip(test_ann, COLORS):
                t_ann, label = t_ann
                t_pr = transform_gt_into_pr(t_ann, gt_ann)
                stats = froc_point(gt_ann, t_pr, .5, use_iou, iou_thres)
                _lls_accuracy, _nlls_per_image = calc_scores(stats, {}, {})
                ax.plot(
                    _nlls_per_image[cat_id][0],
                    _lls_accuracy[cat_id][0],
                    'D',
                    markersize=15,
                    markeredgewidth=3,
                    label=label +
                    f' (FP/image = {np.round(_nlls_per_image[cat_id][0], 2)})',
                    c=c,
                )
                ins.plot(
                    _nlls_per_image[cat_id][0],
                    _lls_accuracy[cat_id][0],
                    'D',
                    markersize=12,
                    markeredgewidth=2,
                    label=label +
                    f' (FP/image = {np.round(_nlls_per_image[cat_id][0], 2)})',
                    c=c,
                )
                ax.hlines(
                    y=_lls_accuracy[cat_id][0],
                    xmin=min_nlls,
                    xmax=max_nlls,
                    linestyles='dashed',
                    colors=c,
                )
                ins.hlines(
                    y=_lls_accuracy[cat_id][0],
                    xmin=min_nlls,
                    xmax=max_nlls,
                    linestyles='dashed',
                    colors=c,
                )

                if bounds is not None:
                    min_rec, _ = bounds[0], bounds[1]
                    if _nlls_per_image[cat_id][0] < min_rec:
                        continue
                else:
                    ax.text(
                        x=_nlls_per_image[cat_id][0], y=_lls_accuracy[cat_id][0],
                        s=f' FP/image = {np.round(_nlls_per_image[cat_id][0], 2)}',
                        fontdict={'fontsize': 40, 'fontweight': 'bold'},
                    )

    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    ax.legend(
        loc='lower left', bbox_to_anchor=(.1, .1),
        fancybox=True, shadow=True, ncol=1, fontsize=45,
    )

    ax.set_title(plot_title, fontdict={'fontsize': 45})
    ax.set_ylabel('Sensitivity', fontdict={'fontsize': 40})
    ax.set_xlabel('FP / image', fontdict={'fontsize': 40})

    ax.tick_params(axis='both', which='major', labelsize=40)
    ins.tick_params(axis='both', which='major', labelsize=35)

    if bounds is not None:
        x_min, x_max, y_min, y_max = bounds
        ax.set_ylim([y_min, y_max])
        ax.set_xlim([x_min, x_max])
    else:
        ax.set_ylim(bottom=0.05, top=1.02)
        
    if csv_path is not None:
        min_x, max_x = min(alternative_nll_high), max(alternative_nll_high)        
        ax.set_xlim([min_x, max_x])

    ax.grid(True, which='both', axis='both', alpha=.5, linestyle='--')

    fig.tight_layout()
    fig.savefig(fname=plot_output_path, dpi=150)

    os.remove('/tmp/tmp_bootstrap_gt.json')
    os.remove('/tmp/tmp_bootstrap_pred.json')
